@@ -2,7 +2,7 @@
 
 {- |
    Module     : Text.XML.HXT.Arrow.Edit
-   Copyright  : Copyright (C) 2006-9 Uwe Schmidt
+   Copyright  : Copyright (C) 2011 Uwe Schmidt
    License    : MIT
 
    Maintainer : Uwe Schmidt (uwe@fh-wedel.de)
@@ -24,8 +24,8 @@ module Text.XML.HXT.Arrow.Edit
 
     , xshowEscapeXml
 
-    , escapeXmlDoc
-    , escapeHtmlDoc
+    , escapeXmlRefs
+    , escapeHtmlRefs
 
     , haskellRepOfXmlDoc
     , treeRepOfXmlDoc
@@ -65,19 +65,21 @@ import           Control.Arrow.ArrowList
 import           Control.Arrow.ArrowIf
 import           Control.Arrow.ArrowTree
 import           Control.Arrow.ListArrow
+import           Control.Arrow.NTreeEdit
 
 import           Data.Char.Properties.XMLCharProps      ( isXmlSpaceChar )
 
 import           Text.XML.HXT.Arrow.XmlArrow
 import           Text.XML.HXT.DOM.Interface
-import qualified Text.XML.HXT.DOM.XmlNode       as XN
+import qualified Text.XML.HXT.DOM.XmlNode               as XN
+import qualified Text.XML.HXT.DOM.ShowXml               as XS
 import           Text.XML.HXT.DOM.FormatXmlTree         ( formatXmlTree )
 import           Text.XML.HXT.Parser.HtmlParsec         ( emptyHtmlTags )
 import           Text.XML.HXT.Parser.XmlEntities        ( xmlEntities )
 import           Text.XML.HXT.Parser.XhtmlEntities      ( xhtmlEntities )
 
 import           Data.List                              ( isPrefixOf )
-import qualified Data.Map                       as M
+import qualified Data.Map                               as M
 import           Data.Maybe
 
 -- ------------------------------------------------------------
@@ -103,29 +105,69 @@ import           Data.Maybe
 
 canonicalizeTree'       :: LA XmlTree XmlTree -> LA XmlTree XmlTree
 canonicalizeTree' toBeRemoved
-    = processChildren (none `when` isText)
+    = processChildren
+      ( (none `when` (isText <+> isXmlPi))      -- remove XML PI and all text around XML root element
+        >>>
+        (deep isPi `when` isDTD)                -- remove DTD parts, except PIs whithin DTD
+      )
       >>>
-      processBottomUp canonicalize1Node
-      where
-      canonicalize1Node :: LA XmlTree XmlTree
-      canonicalize1Node
-          = (deep isPi `when` isDTD)            -- remove DTD parts, except PIs
-            >>>
-            (none `when` toBeRemoved)           -- remove uninteresting nodes
-            >>>
-            ( processAttrl ( processChildren transfCharRef
-                             >>>
-                             collapseXText
-                           )
-              `when` isElem
-            )
-            >>>
-            transfCdata                         -- CDATA -> text
-            >>>
-            transfCharRef                       -- Char refs -> text
-            >>>
-            collapseXText                       -- combine text
+      canonicalizeNodes toBeRemoved
+{-
+canonicalize1Node       :: LA XmlTree XmlTree -> LA XmlTree XmlTree
+canonicalize1Node toBeRemoved
+    = choiceA
+      [ toBeRemoved     :-> none
+      , isElem          :-> ( processAttrl
+                              ( processChildren transfCharRef
+                                >>>
+                                collapseXText'  -- combine text in attribute values
+                              )
+                              >>>
+                              collapseXText'    -- combine text in content
+                            )
+      , isCharRef       :-> ( getCharRef
+                              >>>
+                              arr (\ i -> [toEnum i])
+                              >>>
+                              mkText
+                            )
+      , isCdata         :-> ( getCdata
+                              >>>
+                              mkText
+                            )
+      , this            :-> this
+      ]
+-}
+canonicalizeNodes       :: LA XmlTree XmlTree -> LA XmlTree XmlTree
+canonicalizeNodes toBeRemoved
+    = editNTreeA $
+      [ toBeRemoved     :-> none
+      , ( isElem >>> getAttrl >>> getChildren >>> isCharRef )   -- canonicalize attribute list
+                        :-> ( processAttrl
+                              ( processChildren transfCharRef
+                                >>>
+                                collapseXText'                  -- combine text in attribute values
+                              )
+                              >>>
+                              ( collapseXText'                  -- and combine text in content
+                                `when`
+                                (getChildren >>. has2XText)
+                              )
+                            )
+      , ( isElem >>> (getChildren >>. has2XText) )
+                        :-> collapseXText'                      -- combine text in content
 
+      , isCharRef       :-> ( getCharRef
+                              >>>
+                              arr (\ i -> [toEnum i])
+                              >>>
+                              mkText
+                            )
+      , isCdata         :-> ( getCdata
+                              >>>
+                              mkText
+                            )
+      ]
 
 -- |
 -- Applies some "Canonical XML" rules to a document tree.
@@ -147,12 +189,9 @@ canonicalizeTree' toBeRemoved
 --  - Special characters in attribute values and character content are replaced by character references
 
 canonicalizeAllNodes    :: ArrowList a => a XmlTree XmlTree
-canonicalizeAllNodes
-    = fromLA $
-      canonicalizeTree' ( isCmt         -- remove comment
-                          <+>
-                          isXmlPi       -- remove xml declaration
-                        )
+canonicalizeAllNodes    = fromLA $
+                          canonicalizeTree' isCmt                       -- remove comment
+{-# INLINE canonicalizeAllNodes #-}
 
 -- |
 -- Canonicalize a tree for XPath
@@ -161,9 +200,8 @@ canonicalizeAllNodes
 -- see 'canonicalizeAllNodes'
 
 canonicalizeForXPath    :: ArrowList a => a XmlTree XmlTree
-canonicalizeForXPath
-    = fromLA $
-      canonicalizeTree' isXmlPi
+canonicalizeForXPath    = fromLA $ canonicalizeTree' none               -- comment remains there
+{-# INLINE canonicalizeForXPath #-}
 
 -- |
 -- Canonicalize the contents of a document
@@ -175,27 +213,20 @@ canonicalizeForXPath
 -- see 'canonicalizeAllNodes'
 
 canonicalizeContents    :: ArrowList a => a XmlTree XmlTree
-canonicalizeContents
-    = fromLA $
-      processBottomUp canonicalize1Node
-      where
-      canonicalize1Node :: LA XmlTree XmlTree
-      canonicalize1Node
-          = ( processAttrl ( processChildren transfCharRef
-                             >>>
-                             collapseXText
-                           )
-              `when` isElem
-            )
-            >>>
-            transfCdata                         -- CDATA -> text
-            >>>
-            transfCharRef                       -- Char refs -> text
-            >>>
-            collapseXText                       -- combine text
+canonicalizeContents    = fromLA $
+                          canonicalizeNodes none
+{-# INLINE canonicalizeContents #-}
 
 -- ------------------------------------------------------------
 
+has2XText               :: XmlTrees -> XmlTrees
+has2XText ts0@(t1 : ts1@(t2 : ts2))
+    | XN.isText t1      = if XN.isText t2
+                          then ts0
+                          else has2XText ts2
+    | otherwise         = has2XText ts1
+has2XText _             = []
+
 collapseXText'          :: LA XmlTree XmlTree
 collapseXText'
     = replaceChildren ( listA getChildren >>> arrL (foldr mergeText' []) )
@@ -217,9 +248,7 @@ collapseXText'
 -- This is useful, e.g. after char and entity reference substitution
 
 collapseXText           :: ArrowList a => a XmlTree XmlTree
-collapseXText
-    = fromLA $
-      collapseXText'
+collapseXText           = fromLA collapseXText'
 
 -- |
 -- Applies collapseXText recursively.
@@ -228,9 +257,7 @@ collapseXText
 -- see also : 'collapseXText'
 
 collapseAllXText        :: ArrowList a => a XmlTree XmlTree
-collapseAllXText
-    = fromLA $
-      processBottomUp collapseXText'
+collapseAllXText        = fromLA $ processBottomUp collapseXText'
 
 -- ------------------------------------------------------------
 
@@ -242,10 +269,10 @@ collapseAllXText
 --
 -- So the following law holds
 --
--- > xshowEscape f >>> xread == f
+-- > xshowEscapeXml f >>> xread == f
 
 xshowEscapeXml          :: ArrowXml a => a n XmlTree -> a n String
-xshowEscapeXml f        = xshow (f >>> escapeXmlDoc)
+xshowEscapeXml f        = f >. (uncurry XS.xshow'' escapeXmlRefs)
 
 -- ------------------------------------------------------------
 
@@ -258,131 +285,72 @@ type EntityRefTable     = M.Map Int String
 xmlEntityRefTable
  , xhtmlEntityRefTable  :: EntityRefTable
 
-xmlEntityRefTable   = buildEntityRefTable $ xmlEntities
-xhtmlEntityRefTable = buildEntityRefTable $ xhtmlEntities
+xmlEntityRefTable       = buildEntityRefTable $ xmlEntities
+xhtmlEntityRefTable     = buildEntityRefTable $ xhtmlEntities
 
 buildEntityRefTable     :: [(String, Int)] -> EntityRefTable
 buildEntityRefTable     = M.fromList . map (\ (x,y) -> (y,x) )
 
-escapeText''    :: (Char -> XmlTree) -> (Char -> Bool) -> XmlTree -> XmlTrees
-escapeText'' escChar isEsc t
-    = maybe [t] escape' . XN.getText $ t
-    where
-    escape' "" = [t]            -- empty text nodes remain empty text nodes
-    escape' s  = escape s       -- they do not disapear
-
-    escape ""
-        = []
-    escape (c:s1)
-        | isEsc c
-            = escChar c : escape s1
-    escape s
-        = XN.mkText s1 : escape s2
-        where
-        (s1, s2) = break isEsc s
-
-{-
-escapeCharRef   :: Char -> XmlTree
-escapeCharRef   = XN.mkCharRef . fromEnum
--}
-
-escapeEntityRef :: EntityRefTable -> Char -> XmlTree
-escapeEntityRef entityTable c
-    = maybe (XN.mkCharRef c') XN.mkEntityRef . M.lookup c' $ entityTable
-    where
-    c' = fromEnum c
-
-escapeXmlEntityRef      :: Char -> XmlTree
-escapeXmlEntityRef      = escapeEntityRef xmlEntityRefTable
-
-escapeHtmlEntityRef     :: Char -> XmlTree
-escapeHtmlEntityRef     = escapeEntityRef xhtmlEntityRefTable
-
 -- ------------------------------------------------------------
 
--- |
--- escape all special XML chars into XML entity references or char references
---
--- convert the special XML chars \< and & in text nodes into prefefiened XML entity references,
--- in attribute values also \', \", \>, \\n, \\r and \\t are converted into entity or char references,
--- in comments nothing is converted (see XML standard 2.4, useful e.g. for JavaScript).
-
-escapeXmlDoc            :: ArrowList a => a XmlTree XmlTree
-escapeXmlDoc
-    = fromLA $ escapeDoc escXmlText escXmlAttrValue
+escapeXmlRefs           :: (Char -> String -> String, Char -> String -> String)
+escapeXmlRefs           = (cquote, aquote)
     where
-    escXmlText
-        = arrL $ escapeText'' escapeXmlEntityRef (`elem` "<&")          -- no escape for ", ' and > required: XML standard 2.4
-    escXmlAttrValue
-        = arrL $ escapeText'' escapeXmlEntityRef (`elem` "<>\"\'&\n\r\t")
+    cquote c
+        | c `elem` "<&" = ('&' :)
+                          . ((lookupRef c xmlEntityRefTable) ++)
+                          . (';' :)
+        | otherwise     = (c :)
+    aquote c
+        | c `elem` "<>\"\'&\n\r\t"
+                        = ('&' :)
+                          . ((lookupRef c xmlEntityRefTable) ++)
+                          . (';' :)
+        | otherwise     = (c :)
 
-
--- |
--- escape all special HTML chars into XHTML entity references or char references
---
--- convert the special XML chars \< and & and all none ASCII chars in text nodes into prefefiened XML or XHTML entity references,
--- in attribute values also \', \", \>, \\n, \\r and \\t are converted into entity or char references,
--- in comments nothing is converted
-
--- ------------------------------------------------------------
-
-escapeHtmlDoc           :: ArrowList a => a XmlTree XmlTree
-escapeHtmlDoc
-    = fromLA $ escapeDoc escHtmlText escHtmlAttrValue
+escapeHtmlRefs          :: (Char -> String -> String, Char -> String -> String)
+escapeHtmlRefs          = (cquote, aquote)
     where
-    escHtmlText
-        = arrL $ escapeText'' escapeHtmlEntityRef isHtmlTextEsc
-    escHtmlAttrValue
-        = arrL $ escapeText'' escapeHtmlEntityRef isHtmlAttrEsc
+    cquote c
+        | isHtmlTextEsc c
+                        = ('&' :)
+                          . ((lookupRef c xhtmlEntityRefTable) ++)
+                          . (';' :)
+        | otherwise     = (c :)
+    aquote c
+        | isHtmlAttrEsc c
+                        = ('&' :)
+                          . ((lookupRef c xhtmlEntityRefTable) ++)
+                          . (';' :)
+        | otherwise     = (c :)
 
-    isHtmlTextEsc c
-        = c >= toEnum(128) || ( c `elem` "<&" )
-    isHtmlAttrEsc c
-        = c >= toEnum(128) || ( c `elem` "<>\"\'&\n\r\t" )
+    isHtmlTextEsc c     = c >= toEnum(128) || ( c `elem` "<&" )
+    isHtmlAttrEsc c     = c >= toEnum(128) || ( c `elem` "<>\"\'&\n\r\t" )
 
--- ------------------------------------------------------------
-
-escapeDoc               :: LA XmlTree XmlTree -> LA XmlTree XmlTree -> LA XmlTree XmlTree
-escapeDoc escText escAttr
-    = escape
-    where
-    escape
-        = choiceA
-          [ isElem  :-> ( processChildren escape
-                          >>>
-                          processAttrl escVal
-                        )
-          , isText  :-> escText
-          -- , isCmt   :-> escCmt
-          , isDTD   :-> processTopDown escDTD
-          , this    :-> this
-          ]
-    escVal   = processChildren escAttr
-    escDTD   = escVal `when` ( isDTDEntity <+> isDTDPEntity )
+lookupRef               :: Char -> EntityRefTable -> String
+lookupRef c             = fromMaybe ('#' : show (fromEnum c))
+                          . M.lookup (fromEnum c)
+{-# INLINE lookupRef #-}
 
 -- ------------------------------------------------------------
 
 preventEmptyElements    :: ArrowList a => [String] -> Bool -> a XmlTree XmlTree
 preventEmptyElements ns isHtml
-    = fromLA $ insertDummyElem
+    = fromLA $
+      editNTreeA [ ( isElem
+                     >>>
+                     isNoneEmpty
+                     >>>
+                     neg getChildren
+                   )
+                   :-> replaceChildren (txt "")
+                 ]
     where
     isNoneEmpty
         | not (null ns) = hasNameWith (localPart >>> (`elem` ns))
         | isHtml        = hasNameWith (localPart >>> (`notElem` emptyHtmlTags))
         | otherwise     = this
 
-    insertDummyElem
-        = processBottomUp
-          ( replaceChildren (txt "")
-            `when`
-            ( isElem
-              >>>
-              isNoneEmpty
-              >>>
-              neg getChildren
-            )
-          )
-
 -- ------------------------------------------------------------
 
 -- |
@@ -441,25 +409,19 @@ addHeadlineToXmlDoc
 
 -- ------------------------------------------------------------
 
-removeComment'          :: LA XmlTree XmlTree
-removeComment'          = none `when` isCmt
-
 -- |
--- remove Comments: @none `when` isCmt@
+-- remove a Comment node
 
 removeComment           :: ArrowXml a => a XmlTree XmlTree
-removeComment           = fromLA $ removeComment'
+removeComment           = none `when` isCmt
 
 -- |
--- remove all comments recursively
+-- remove all comments in a tree recursively
 
 removeAllComment        :: ArrowXml a => a XmlTree XmlTree
-removeAllComment        = fromLA $ processBottomUp removeComment'
+removeAllComment        = fromLA $ editNTreeA [isCmt :-> none]
 
--- ----------
-
-removeWhiteSpace'       :: LA XmlTree XmlTree
-removeWhiteSpace'       = none `when` isWhiteSpace
+-- ------------------------------------------------------------
 
 -- |
 -- simple filter for removing whitespace.
@@ -470,7 +432,7 @@ removeWhiteSpace'       = none `when` isWhiteSpace
 -- see also : 'removeAllWhiteSpace', 'removeDocWhiteSpace'
 
 removeWhiteSpace        :: ArrowXml a => a XmlTree XmlTree
-removeWhiteSpace        = fromLA $ removeWhiteSpace'
+removeWhiteSpace        = fromLA $ none `when` isWhiteSpace
 
 -- |
 -- simple recursive filter for removing all whitespace.
@@ -481,7 +443,10 @@ removeWhiteSpace        = fromLA $ removeWhiteSpace'
 -- see also : 'removeWhiteSpace', 'removeDocWhiteSpace'
 
 removeAllWhiteSpace     :: ArrowXml a => a XmlTree XmlTree
-removeAllWhiteSpace     = fromLA $ processBottomUp removeWhiteSpace'
+removeAllWhiteSpace     = fromLA $ editNTreeA [isWhiteSpace :-> none]
+                       -- fromLA $ processBottomUp removeWhiteSpace'    -- less efficient
+
+-- ------------------------------------------------------------
 
 -- |
 -- filter for removing all not significant whitespace.
@@ -673,43 +638,33 @@ insertNothing _         = none
 
 -- ------------------------------------------------------------
 
-transfCdata'            :: LA XmlTree XmlTree
-transfCdata'            = (getCdata >>> mkText) `when` isCdata
-
 -- |
--- converts a CDATA section node into a normal text node
+-- converts a CDATA section into normal text nodes
 
 transfCdata             :: ArrowXml a => a XmlTree XmlTree
 transfCdata             = fromLA $
-                          transfCdata'
+                          (getCdata >>> mkText) `when` isCdata
 
 -- |
 -- converts CDATA sections in whole document tree into normal text nodes
 
 transfAllCdata          :: ArrowXml a => a XmlTree XmlTree
-transfAllCdata          = fromLA $
-                          processBottomUp transfCdata'
-
---
-
-transfCharRef'          :: LA XmlTree XmlTree
-transfCharRef'          = ( getCharRef >>> arr (\ i -> [toEnum i]) >>> mkText )
-                          `when`
-                          isCharRef
+transfAllCdata          = fromLA $ editNTreeA [isCdata :-> (getCdata >>> mkText)]
 
 -- |
--- converts character references to normal text
+-- converts a character reference to normal text
 
 transfCharRef           :: ArrowXml a => a XmlTree XmlTree
 transfCharRef           = fromLA $
-                          transfCharRef'
+                          ( getCharRef >>> arr (\ i -> [toEnum i]) >>> mkText )
+                          `when`
+                          isCharRef
 
 -- |
 -- recursively converts all character references to normal text
 
 transfAllCharRef        :: ArrowXml a => a XmlTree XmlTree
-transfAllCharRef        = fromLA $
-                          processBottomUp transfCharRef'
+transfAllCharRef        = fromLA $ editNTreeA [isCharRef :-> (getCharRef >>> arr (\ i -> [toEnum i]) >>> mkText)]
 
 -- ------------------------------------------------------------
 
@@ -741,7 +696,7 @@ addDefaultDTDecl
             <+>
             txt "\n"
             <+>
-            ( getChildren >>> (none `when` isDTDDoctype) )      -- remove old DTD stuff
+            ( getChildren >>> (none `when` isDTDDoctype) )      -- remove old DTD decl
           )
 
 -- ------------------------------------------------------------
@@ -762,7 +717,7 @@ hasXmlPi
 addXmlPi                :: ArrowXml a => a XmlTree XmlTree
 addXmlPi
     = fromLA
-      ( insertChildrenAt 0 ( ( mkPi (mkSNsName t_xml) none
+      ( insertChildrenAt 0 ( ( mkPi (mkName t_xml) none
                                >>>
                                addAttr a_version "1.0"
                              )
