@@ -4,14 +4,14 @@
 
 {- |
    Module     : Text.XML.HXT.DOM.QualifiedName
-   Copyright  : Copyright (C) 2008 Uwe Schmidt
+   Copyright  : Copyright (C) 2011 Uwe Schmidt
    License    : MIT
 
    Maintainer : Uwe Schmidt (uwe@fh-wedel.de)
    Stability  : stable
    Portability: portable
 
-   The core data types of the HXT DOM.
+   The types and functions for qualified names
 
 -}
 
@@ -19,7 +19,7 @@
 
 module Text.XML.HXT.DOM.QualifiedName
     ( QName
-    , XName
+    , XName(unXN)
     , NsEnv
 
     , mkQName
@@ -39,6 +39,7 @@ module Text.XML.HXT.DOM.QualifiedName
     , newXName
     , nullXName
     , isNullXName
+    , newQName
 
     , mkQName'
     , namePrefix'
@@ -50,6 +51,7 @@ module Text.XML.HXT.DOM.QualifiedName
     , setNamespaceUri'
 
     , qualifiedName
+    , qualifiedName'
     , universalName
     , universalUri
     , buildUniversalName
@@ -84,6 +86,7 @@ import           Control.Arrow                  ( (***) )
 
 import           Control.Concurrent.MVar
 import           Control.DeepSeq
+import           Control.FlatSeq
 
 import           Data.AssocList
 import           Data.Binary
@@ -110,7 +113,44 @@ import Data.Char.Properties.XMLCharProps        ( isXmlNCNameStartChar
 -- Names are always reduced to normal form, and they are stored internally in a name cache
 -- for sharing equal names by the same data structure
 
-type XName      = Atom
+data XName                      = XN { _idXN    :: ! Int        -- for optimization of equality test, see Eq instance
+                                     ,  unXN    ::   String
+                                     }
+                                  deriving (Typeable)
+
+instance Eq XName where
+    (XN id1 _) == (XN id2 _)    = id1 == id2
+
+instance Ord XName where
+    compare (XN _ n1) (XN _ n2) = compare n1 n2
+{-
+instance Read XName where
+    readsPrec p str             = [ (newXName x, y) | (x, y) <- readsPrec p str ]
+
+instance Show XName where
+    show (XN _ s)               = show s
+-}
+instance NFData XName where
+    rnf (XN _ s)                = rnf s
+
+instance WNFData XName where
+    rwnf (XN _ s)               = rnf s
+
+instance Binary XName where
+    put (XN _ s)                = put s
+    get                         = do
+                                  s <- get
+                                  return $! newXName s
+
+-----------------------------------------------------------------------------
+
+-- |
+-- Type for the namespace association list, used when propagating namespaces by
+-- modifying the 'QName' values in a tree
+
+type NsEnv              = AssocList XName XName
+
+-----------------------------------------------------------------------------
 
 -- |
 -- Namespace support for element and attribute names.
@@ -121,120 +161,81 @@ type XName      = Atom
 -- When dealing with namespaces, the document tree must be processed by 'Text.XML.HXT.Arrow.Namespace.propagateNamespaces'
 -- to split names of structure \"prefix:localPart\" and label the name with the apropriate namespace uri
 
-data QName      = LP ! XName
-                | PX ! XName ! QName
-                | NS ! XName ! QName
-             deriving (Ord, Show, Read, Typeable)
-
--- |
--- Type for the namespace association list, used when propagating namespaces by
--- modifying the 'QName' values in a tree
-
-type NsEnv = AssocList XName XName
+data QName      = QN { localPart'       :: ! XName
+                     , namePrefix'      :: ! XName
+                     , namespaceUri'    :: ! XName
+                     }
+             deriving (Typeable)
 
 -- -----------------------------------------------------------------------------
 
 -- | Two QNames are equal if (1. case) namespaces are both empty and the qualified names
 -- (prefix:localpart) are the same or (2. case) namespaces are set and namespaces and
--- local parts both are equal
+-- local parts are equal
 
 instance Eq QName where
-    (LP lp1)     == (LP lp2)            = lp1 == lp2
-    (PX px1 qn1) == (PX px2 qn2)        = px1 == px2 && qn1== qn2
-    (NS ns1 qn1) == (NS ns2 qn2)        = ns1 == ns2 && localPart' qn1 == localPart' qn2
-    n1@(PX _ _)  == n2@(LP _)           = qualifiedName n1 == qualifiedName n2
-    n1@(LP _)    == n2@(PX _ _)         = qualifiedName n1 == qualifiedName n2
-    _            == _                   = False
+    (QN lp1 px1 ns1) == (QN lp2 px2 ns2)
+        | ns1 /= ns2            = False                 -- namespaces are set and differ
+        | not (isNullXName ns1) = lp1 == lp2            -- namespaces are set and are equal: local parts must be equal
+        | otherwise             = lp1 == lp2            -- no namespaces are set: local parts must be equal
+                                  &&                    -- and prefixes are not set or they are equal
+                                  px1 == px2
 
--- the 4. and 5. rule are only neccessary when someone
--- uses XML names not in a systematical way
--- and does things like "mkName("x:y") == mkPrefixLocalPart("x","y")
+instance NFData  QName
+instance WNFData QName
 
-instance NFData QName where
+instance Show QName where
+    show                        = showQN
+
+-- -----------------------------------------------------------------------------
 
 instance Binary QName where
-    put qn              = let
-                          px = namePrefix   qn
-                          lp = localPart    qn
-                          ns = namespaceUri qn
-                          in
-                          put px >> put lp >> put ns
+    put (QN lp px ns)   = put (unXN px) >>
+                          put (unXN lp) >>
+                          put (unXN ns)
     get                 = do
                           px <- get
                           lp <- get
                           ns <- get
-                          return $ mkQName px lp ns
+                          return $! newNsName lp px ns
+                          --     ^^
+                          -- strict apply !!!
+                          -- build the QNames strict, else the name sharing optimization will not be in effect
 
 -- -----------------------------------------------------------------------------
 
-newXName                :: String -> XName
-newXName                = newAtom
-
 isNullXName             :: XName -> Bool
 isNullXName             = (== nullXName)
-
-nullXName               :: XName
-nullXName               = newXName ""
-
--- | access name prefix
-
-namePrefix'             :: QName -> XName
-namePrefix' (LP _)      = nullXName
-namePrefix' (PX px _)   = px
-namePrefix' (NS _ n)    = namePrefix' n
-
--- | access local part
-
-localPart'              :: QName -> XName
-localPart' (LP lp)      = lp
-localPart' (PX _ n)     = localPart' n
-localPart' (NS _ n)     = localPart' n
-
--- | access namespace uri
-
-namespaceUri'           :: QName -> XName
-namespaceUri' (NS ns _) = ns
-namespaceUri' _         = nullXName
+{-# INLINE isNullXName #-}
 
 namePrefix              :: QName -> String
-namePrefix              = show . namePrefix'
+namePrefix              = unXN . namePrefix'
+{-# INLINE namePrefix #-}
 
 localPart               :: QName -> String
-localPart               = show . localPart'
+localPart               = unXN . localPart'
+{-# INLINE localPart #-}
 
 namespaceUri            :: QName -> String
-namespaceUri            = show . namespaceUri'
+namespaceUri            = unXN . namespaceUri'
+{-# INLINE namespaceUri #-}
 
 -- ------------------------------------------------------------
 
 -- | set name prefix
 
-setNamespaceUri'                :: XName -> QName -> QName
-setNamespaceUri' ns (NS _ n)    = if isNullXName ns
-                                  then n
-                                  else NS ns n
-setNamespaceUri' ns n           = if isNullXName ns
-                                  then n
-                                  else NS ns n
+setNamespaceUri'                        :: XName -> QName -> QName
+setNamespaceUri' ns (QN lp px _ns)      = newQName lp px ns
 
 -- | set local part
 
-setLocalPart'                   :: XName -> QName -> QName
-setLocalPart' lp (LP _)         = LP lp
-setLocalPart' lp (PX px n)      = PX px (setLocalPart' lp n)
-setLocalPart' lp (NS ns n)      = NS ns (setLocalPart' lp n)
+setLocalPart'                           :: XName -> QName -> QName
+setLocalPart' lp (QN _lp px ns)         = newQName lp px ns
 
 -- | set name prefix
 
-setNamePrefix'                  :: XName -> QName -> QName
-setNamePrefix' px (PX _ n)      = if px == nullXName
-                                  then n
-                                  else PX px n
-setNamePrefix' px n@(LP _)      = if px == nullXName
-                                  then n
-                                  else PX px n
-setNamePrefix' px (NS ns n)     = NS ns (setNamePrefix' px n)
-
+setNamePrefix'                          :: XName -> QName -> QName
+setNamePrefix' px (QN lp _px ns)        = newQName lp px ns
 
 -- ------------------------------------------------------------
 
@@ -242,9 +243,16 @@ setNamePrefix' px (NS ns n)     = NS ns (setNamePrefix' px n)
 -- builds the full name \"prefix:localPart\", if prefix is not null, else the local part is the result
 
 qualifiedName                   :: QName -> String
-qualifiedName (LP lp)           = show lp
-qualifiedName (PX px n)         = show px ++ (':' : qualifiedName n)
-qualifiedName (NS _ n)          = qualifiedName n
+qualifiedName (QN lp px _ns)
+    | isNullXName px            = unXN lp
+    | otherwise                 = unXN px ++ (':' : unXN lp)
+
+-- | functional list version of qualifiedName used in xshow
+
+qualifiedName'                   :: QName -> String -> String
+qualifiedName' (QN lp px _ns)
+    | isNullXName px            = (unXN lp ++)
+    | otherwise                 = (unXN px ++) . (':' :) . (unXN lp ++)
 
 -- |
 -- builds the \"universal\" name, that is the namespace uri surrounded with \"{\" and \"}\" followed by the local part
@@ -266,21 +274,24 @@ universalUri                    = buildUniversalName (++)
 -- namespace uri and local part are combined with the combining function given by the first parameter
 
 buildUniversalName              :: (String -> String -> String) -> QName -> String
-buildUniversalName bf (NS ns n) = show ns `bf` localPart n
-buildUniversalName _  n         = localPart n
+buildUniversalName bf n@(QN _lp _px ns)
+    | isNullXName ns            = localPart n
+    | otherwise                 = unXN ns `bf` localPart n
+
+showQN                          :: QName -> String
+showQN n
+    | null ns                   = show $ qualifiedName n
+    | otherwise                 = show $ "{" ++ ns ++ "}" ++ qualifiedName n
+    where
+    ns = namespaceUri n
 
 -- ------------------------------------------------------------
 --
 -- internal XName functions
 
 mkQName'                        :: XName -> XName -> XName -> QName
-mkQName' px lp ns
-    | isNullXName ns            =       px_lp
-    | otherwise                 = NS ns px_lp
-    where
-    px_lp
-        | isNullXName px        = LP lp
-        | otherwise             = PX px (LP lp)
+mkQName' px lp ns               = newQName lp px ns
+{-# DEPRECATED mkQName' "use newQName instead with lp px ns param seq " #-}
 
 -- ------------------------------------------------------------
 
@@ -291,10 +302,8 @@ mkQName' px lp ns
 
 mkPrefixLocalPart               :: String -> String -> QName
 mkPrefixLocalPart px lp
-    | null px                   =                  n1
-    | otherwise                 = PX (newXName px) n1
-    where
-    n1 = LP (newXName lp)
+    | null px                   = newLpName lp
+    | otherwise                 = newPxName lp px
 
 -- |
 -- constructs a simple, namespace unaware name.
@@ -307,10 +316,10 @@ mkName n
     | (':' `elem` n)
       &&
       not (null px)                     -- more restrictive: isWellformedQualifiedName n
-                                = mkPrefixLocalPart px lp
-    | otherwise                 = mkPrefixLocalPart "" n
+                                = newPxName lp px
+    | otherwise                 = newLpName n
     where
-    (px, (_:lp)) = span (/= ':') n
+    (px, (_ : lp)) = span (/= ':') n
 
 -- |
 -- constructs a complete qualified name with 'namePrefix', 'localPart' and 'namespaceUri'.
@@ -320,10 +329,8 @@ mkName n
 
 mkQName                         :: String -> String -> String -> QName
 mkQName px lp ns
-    | null ns                   =                  n1
-    | otherwise                 = NS (newXName ns) n1
-    where
-    n1 = mkPrefixLocalPart px lp
+    | null ns                   = mkPrefixLocalPart px lp
+    | otherwise                 = newNsName lp px ns
 
 -- ------------------------------------------------------------
 
@@ -332,6 +339,7 @@ mkQName px lp ns
 
 mkSNsName                       :: String -> QName
 mkSNsName                       = mkName
+{-# DEPRECATED mkSNsName "use mkName instead" #-}
 
 -- |
 -- constructs a simple, namespace aware name, with prefix:localPart as first parameter,
@@ -339,10 +347,18 @@ mkSNsName                       = mkName
 --
 -- see also 'mkName', 'mkPrefixLocalPart'
 
-mkNsName                        :: String -> String -> QName
+{-
+mkNsName                          :: String -> String -> QName
+mkNsName n ns                     = trace ("mkNsName: " ++ show n ++ " " ++ show ns) (mkNsName' n ns)
+-}
+
+mkNsName                          :: String -> String -> QName
 mkNsName n ns
-    | null ns                   =                   mkName n
-    | otherwise                 = NS (newXName ns) (mkName n)
+    | null ns			= qn
+    | otherwise			= setNamespaceUri' ns' qn
+    where
+    qn                          = mkName n
+    ns'				= newXName ns
 
 -- ------------------------------------------------------------
 
@@ -388,27 +404,8 @@ normalizeNsUri                  = map toLower . stripSlash
 -- see 'isWellformedQName' and 'isWellformedQualifiedName' for error checking.
 
 setNamespace                    :: NsEnv -> QName -> QName
-setNamespace env n@(PX px _)    = attachNS env px        n              -- none empty prefix found
-setNamespace env n@(LP _)       = attachNS env nullXName n              -- use default namespace uri
-setNamespace env (NS _ n)       = setNamespace env n
-
-attachNS                        :: NsEnv -> XName -> QName -> QName
-attachNS env px n1              = maybe n1 (\ ns -> NS ns n1) . lookup px $ env
-
-xmlnsNamespaceXName             :: XName
-xmlnsNamespaceXName             = newXName xmlnsNamespace
-
-xmlnsXName                      :: XName
-xmlnsXName                      = newXName a_xmlns
-
-xmlnsQN                         :: QName
-xmlnsQN                         = NS xmlnsNamespaceXName (LP xmlnsXName)
-
-xmlNamespaceXName               :: XName
-xmlNamespaceXName               = newXName xmlNamespace
-
-xmlXName                        :: XName
-xmlXName                        = newXName a_xml
+setNamespace env n@(QN lp px _ns)
+                                = maybe n (\ ns -> newQName lp px ns) . lookup px $ env
 
 -- -----------------------------------------------------------------------------
 --
@@ -440,11 +437,13 @@ isWellformedQualifiedName s
 -- predicate is used in filter 'valdateNamespaces'.
 
 isWellformedQName               :: QName -> Bool
-isWellformedQName (LP lp)       = isNCName . show $ lp                          -- rule [8] XML Namespaces
-isWellformedQName (PX px n)     = (isNCName . show) px                          -- rule [7] XML Namespaces
+isWellformedQName (QN lp px _ns)
+                                = (isNCName . unXN) lp                          -- rule [8] XML Namespaces
                                   &&
-                                  isWellformedQName n
-isWellformedQName (NS _ n)      = isWellformedQName n
+                                  ( isNullXName px
+                                    ||
+                                    (isNCName . unXN) px                        -- rule [7] XML Namespaces
+                                  )
 
 -- |
 -- test whether an attribute name is a namesapce declaration name.
@@ -454,47 +453,44 @@ isWellformedQName (NS _ n)      = isWellformedQName n
 -- predicate is used in filter 'valdateNamespaces'.
 
 isWellformedNSDecl              :: QName -> Bool
-isWellformedNSDecl n            = not (isNameSpaceName n)
+isWellformedNSDecl n
+                                = not (isNameSpaceName n)
                                   ||
                                   isWellformedNameSpaceName n
 
 -- |
 -- test for a namespace name to be well formed
 
-isWellformedNameSpaceName               :: QName -> Bool
-isWellformedNameSpaceName (LP lp)       = lp == xmlnsXName
-isWellformedNameSpaceName (PX px n)     = px == xmlnsXName
-                                          &&
-                                          not (null lp')
-                                          &&
-                                          not (a_xml `isPrefixOf` lp')
-                                          where
-                                          lp' = localPart n
-isWellformedNameSpaceName (NS _ n)      = isWellformedNSDecl n
+isWellformedNameSpaceName       :: QName -> Bool
+isWellformedNameSpaceName n@(QN lp px _ns)
+    | isNullXName px            = lp == xmlnsXName
+    | otherwise                 = px == xmlnsXName
+                                  &&
+                                  not (null lp')
+                                  &&
+                                  not (a_xml `isPrefixOf` lp')
+    where
+    lp'                         = localPart n
+
 
 -- |
 -- test whether a name is a namespace declaration attribute name
 
-isNameSpaceName                 :: QName -> Bool
-isNameSpaceName (LP lp)         = lp == xmlnsXName
-isNameSpaceName (PX px _)       = px == xmlnsXName
-isNameSpaceName (NS _  n)       = isNameSpaceName n
+isNameSpaceName                         :: QName -> Bool
+isNameSpaceName (QN lp px _ns)
+    | isNullXName px                    = lp == xmlnsXName
+    | otherwise                         = px == xmlnsXName
 
 -- |
 --
 -- predicate is used in filter 'valdateNamespaces'.
 
-isDeclaredNamespace             :: QName -> Bool
-isDeclaredNamespace (NS ns n)   = isNS ns        n
-isDeclaredNamespace        n    = isNS nullXName n
-
-isNS                            :: XName -> QName -> Bool
-isNS _  (LP _)                  = True                          -- no namespace used
-isNS ns (PX px _)
-    | px == xmlnsXName          = ns == xmlnsNamespaceXName     -- "xmlns" has a predefined namespace uri
-    | px == xmlXName            = ns == xmlNamespaceXName       -- "xml" has a predefiend namespace"
-    | otherwise                 = ns /= nullXName               -- namespace values are not empty
-isNS ns (NS _ n)                = isNS ns n                     -- this does not occur, but warning is prevented
+isDeclaredNamespace                     :: QName -> Bool
+isDeclaredNamespace (QN _lp px ns)
+    | isNullXName px                    = True                          -- no namespace used
+    | px == xmlnsXName                  = ns == xmlnsNamespaceXName     -- "xmlns" has a predefined namespace uri
+    | px == xmlXName                    = ns == xmlNamespaceXName       -- "xml" has a predefiend namespace"
+    | otherwise                         = not (isNullXName ns)          -- namespace values are not empty
 
 -- -----------------------------------------------------------------------------
 
@@ -503,50 +499,127 @@ toNsEnv                         = map (newXName *** newXName)
 
 -- -----------------------------------------------------------------------------
 
--- the name cache, same implementation strategy as in Data.Atom,
--- but conversion to and from ByteString prevented
+-- the name and string cache
 
-type Atoms      = M.Map String String
-
-newtype Atom    = A String
-                  deriving (Eq, Ord, Typeable)
+data NameCache          = NC { _newXN    :: ! Int                                       -- next free name id
+                             , _xnCache  :: ! (M.Map String XName)
+                             , _qnCache  :: ! (M.Map (XName, XName, XName) QName)       -- we need another type than QName
+                             }                                                          -- for the key because of the unusable
+                                                                                        -- Eq instance of QName
+type ChangeNameCache r  = NameCache -> (NameCache, r)
 
 -- ------------------------------------------------------------
 
--- | the internal cache for the strings
+-- | the internal cache for QNames (and name strings)
 
-theAtoms        :: MVar Atoms
-theAtoms        = unsafePerformIO (newMVar M.empty)
-{-# NOINLINE theAtoms #-}
+theNameCache            :: MVar NameCache
+theNameCache            = unsafePerformIO (newMVar $ initialCache)
+{-# NOINLINE theNameCache #-}
 
--- | insert a bytestring into the atom cache
+initialXNames           :: [XName]
 
-insertAtom      :: String -> Atoms -> (Atoms, Atom)
-insertAtom s m  = maybe (M.insert {- (trace (show s) s) -} s s m, deepseq s (A s))
-                        (\ s' -> (m, A s'))
-                  .
-                  M.lookup s $ m
+nullXName
+ , xmlnsNamespaceXName
+ , xmlnsXName
+ , xmlNamespaceXName
+ , xmlXName             :: XName
 
--- | creation of an @Atom@ from a @String@
+initialXNames@
+ [ nullXName
+ , xmlnsNamespaceXName
+ , xmlnsXName
+ , xmlNamespaceXName
+ , xmlXName
+ ]                      = zipWith XN [0..] $
+                          [ ""
+                          , xmlnsNamespace
+                          , a_xmlns
+                          , xmlNamespace
+                          , a_xml
+                          ]
 
-newAtom         :: String -> Atom
-newAtom         = unsafePerformIO . newAtom'
-{-# NOINLINE newAtom #-}
+initialQNames           :: [QName]
 
--- | The internal operation running in the IO monad
-newAtom'        :: String -> IO Atom
-newAtom' s      = do
-                  m <- takeMVar theAtoms
-                  let (m', a) = insertAtom s m
-                  putMVar theAtoms m'
-                  return a
+xmlnsQN                 :: QName
 
-instance Read Atom where
-    readsPrec p str = [ (newAtom x, y) | (x, y) <- readsPrec p str ]
+initialQNames@
+ [xmlnsQN]              = [QN xmlnsXName nullXName xmlnsNamespaceXName]
 
-instance Show Atom where
-    show (A s)  = s
+initialCache            :: NameCache
+initialCache            = NC
+                          (length initialXNames)
+                          (M.fromList $ map (\ xn -> (unXN xn, xn)) initialXNames)
+                          (M.fromList $ map (\ qn@(QN lp px ns) -> ((lp, px, ns), qn)) initialQNames)
 
-instance NFData Atom where
+-- ------------------------------------------------------------
+
+changeNameCache         :: NFData r => ChangeNameCache r -> r
+changeNameCache action  = unsafePerformIO changeNameCache'
+    where
+    changeNameCache'    = do
+                          -- putStrLn "takeMVar"
+                          c <- takeMVar theNameCache
+                          let (c', res) = action c
+                          c' `seq` -- rnf res `seq`
+                             putMVar theNameCache c'
+                          -- putStrLn "putMVar"
+                          return res
+
+{-# NOINLINE changeNameCache #-}
+
+newXName'               :: String -> ChangeNameCache XName
+newXName' n c@(NC nxn xm qm)
+                        = case M.lookup n xm of
+                          Just xn       -> (c, xn)
+                          Nothing       -> let nxn' = nxn + 1 in
+                                           let xn   = (XN nxn n) in
+                                           let xm'  = M.insert n xn xm in
+                                           -- trace ("newXName: XN " ++ show nxn ++ " " ++ show n) $
+                                           rnf xn `seq` (NC nxn' xm' qm, xn)
+
+newQName'               :: XName -> XName -> XName -> ChangeNameCache QName
+newQName' lp px ns c@(NC nxn xm qm)
+                        = case M.lookup q' qm of
+                          Just qn       -> -- trace ("oldQName: " ++ show qn) $                 -- log evaluation sequence
+                                           (c, qn)
+                          Nothing       -> let qm'  = M.insert q' q qm in
+                                           -- trace ("newQName: " ++ show q) $                  -- log insertion of a new QName
+                                           q `seq` (NC nxn xm qm', q)
+    where
+    q'                  = (lp, px, ns)
+    q                   = QN lp px ns
+
+andThen                 :: ChangeNameCache r1 ->
+                           (r1 -> ChangeNameCache r2) -> ChangeNameCache r2
+andThen a1 a2 c0        = let (c1, r1) = a1 c0 in
+                          (a2 r1) c1
+
+newXName                :: String -> XName
+newXName n              = changeNameCache $
+                          newXName' n
+
+newQName                :: XName -> XName -> XName -> QName
+newQName lp px ns       = lp `seq` px `seq` ns `seq`                    -- XNames must be evaluated, else MVar blocks
+                          ( changeNameCache $
+                            newQName' lp px ns
+                          )
+
+newLpName               :: String -> QName
+newLpName lp            = changeNameCache $
+                          newXName' lp `andThen` \ lp' ->
+                          newQName' lp' nullXName nullXName
+
+newPxName               :: String -> String -> QName
+newPxName lp px         = changeNameCache $
+                          newXName' lp `andThen` \ lp' ->
+                          newXName' px `andThen` \ px' ->
+                          newQName' lp' px' nullXName
+
+newNsName               :: String -> String -> String -> QName
+newNsName lp px ns      = changeNameCache $
+                          newXName' lp `andThen` \ lp' ->
+                          newXName' px `andThen` \ px' ->
+                          newXName' ns `andThen` \ ns' ->
+                          newQName' lp' px' ns'
 
 -----------------------------------------------------------------------------
